from pathlib import Path

APP_DIR = Path().home().joinpath('pydbbackups').absolute()

BACKUPS_DIR = APP_DIR.joinpath('backups')
BACKUPS_DATA_DIR = APP_DIR.joinpath('backups.json')


def build_config():
    APP_DIR.mkdir(exist_ok=True, parents=True)
    BACKUPS_DIR.mkdir(exist_ok=True, parents=True)

    if not BACKUPS_DATA_DIR.exists():
        BACKUPS_DATA_DIR.touch()
        BACKUPS_DATA_DIR.write_text('[]', encoding='utf-8')

    BACKUPS_DATA_DIR.touch(exist_ok=True)



import json
from importlib import import_module
from typing import Generator, List
from pathlib import Path

from pydbbackups.cli.config import BACKUPS_DATA_DIR, BACKUPS_DIR
from pydbbackups import Backup
from pydbbackups.cli.models import BackupFile, BackupData
from datetime import datetime
from pydbbackups.backups import mongodb, postgres

databases = {
    'postgres': 'Postgres',
    'mongodb': 'MongoDB'
}


def get_backups_data() -> List[BackupData]:
    b_data = BACKUPS_DATA_DIR.read_bytes()
    # 2023-04-16 11:55:13.039897
    data = [
        BackupData(
            id=v['id'],
            name=v['name'],
            ext=v['ext'],
            database_name=v['database_name'],
            backup=Path(v['backup']),
            created_at=datetime.fromisoformat(v['created_at'])
        ) for v in json.loads(b_data)
    ]
    return data


def get_backups_files() -> Generator[BackupFile, None, None]:
    dir = BACKUPS_DIR.iterdir()

    for f in dir:
        [id, name] = f.name.split("___")
        [name, ext] = name.split('.')
        yield BackupFile(id=int(id), name=name, ext=ext, file=f)


def get_backup_class(name: str) -> Backup:
    m = import_module(f"pydbbackups.backups.{name}")
    return getattr(m, databases[name])


def save_backup_file(meta: BackupFile, content: bytes | str) -> BackupFile:
    if "___" in meta.name:
        raise ValueError("the backup name should not include <___>")

    path = BACKUPS_DIR.joinpath(
        f"{meta.id}___{meta.name}{f'.{meta.ext}' if meta.ext else ''}")
    path.touch()
    if isinstance(content, bytes):
        path.write_bytes(content)
    else:
        path.write_text(content)

    return BackupFile(id=meta.id, name=meta.name, ext=meta.ext, file=path)


class DTEncoder(json.JSONEncoder):
    """ Datetime encoder """

    def default(self, obj):
        if isinstance(obj, datetime):
            return str(obj.isoformat())
        if isinstance(obj, Path):
            return str(obj.resolve())
        return json.JSONEncoder.default(self, obj)


# --------- Extension formatters ---------

def mongo_ext_formatter(name: str, **kwargs):
    if not kwargs.get('compress', False):
        return f"{name}.{mongodb.DUMP_DEFAULT_FORMAT}"
    else:
        return f"{name}.{mongodb.DUMP_GZIP_FORMAT}"


def postgres_ext_formatter(name, **kwargs):
    dictionary = {
        postgres.DUMP_CUSTOM_FORMAT: 'dump',
        postgres.DUMP_SQL_FORMAT: 'sql',
        postgres.DUMP_TAR_FORMAT: 'tar',
        postgres.DUMP_DIRECTORY_FORMAT: '',
        'DEFAULT': postgres.DUMP_DEFAULT_FORMAT
    }

    format = kwargs.get('format', None)
    if not format:
        ext = dictionary.get(dictionary.get('DEFAULT'))
    else:
        ext = dictionary.get(format)
    return f"{name}{f'.{ext}' if len(ext) > 0 else ''}"



from .base import Backup
import subprocess
from io import BytesIO
from pathlib import Path


DUMP_DEFAULT_FORMAT = "archive"
DUMP_GZIP_FORMAT = "gz"
DUMP_ARCHIVE_FORMAT = "archive"


class MongoDB(Backup):
    CMDS_TO_CHECK = [('mongodump', '--version')]

    def dump(self, **kwargs):
        config = {
            'compress': kwargs.get('compress', False),
        }

        if self.uri:
            args = [
                '-uri', self.uri
            ]
        else:
            args = [
                '--username', self.username,
                '--password', self.password,
                '--host', self.host,
                '--port', f"{self.port}",
                '--db', self.database,
                '--authenticationDatabase', kwargs.get(
                    'authentication_database', 'admin'),
                '--archive'
            ]

        if config['compress']:
            args.append('--gzip')

        p = subprocess.Popen([
            "mongodump",
            *args,
        ], stdout=subprocess.PIPE, stderr=subprocess.PIPE)

        status_code = p.wait()
        if status_code > 0:
            raise Exception(p.stderr.read().decode('utf-8'))

        return BytesIO(p.stdout.read())

    def restore(self, file_path: Path) -> BytesIO:
        args = [
            '--username', self.username,
            '--password', self.password,
            '--host', self.host,
            '--port', f"{self.port}",
            '--nsInclude',
            self.database,
        ]

        if file_path.name.endswith(DUMP_GZIP_FORMAT):
            args.append('--gzip')

        args.append(f'--archive={file_path.resolve()}')

        p = subprocess.Popen([
            "mongorestore",
            *args,
        ], stdout=subprocess.PIPE, stderr=subprocess.PIPE)

        status_code = p.wait()
        if status_code > 0:
            raise Exception(p.stderr.read().decode('utf-8'))

        return BytesIO(p.stdout.read())



class MethodNotImplemented(Exception):
    """ Raised when a one method not implemented is called """

    def __init__(self, name: str = None) -> None:
        message = ""
        if not name:
            message += "Method"
        else:
            message += f"{name} method"

        super().__init__(f"{message} not implemented")


class CommandNotFound(Exception):
    """ Raised when python not found a command """

    def __init__(self, name: str) -> None:
        super().__init__(f'{name} command was not found')



import click
import getpass
import inquirer
import time

from rich.table import Table
from rich.console import Console
from pydbbackups.cli.config import build_config
from pydbbackups.cli.services import BackupsService


@click.group(invoke_without_command=True, no_args_is_help=True)
@click.version_option('0.1.0', prog_name='dbbackups')
def app():
    """ Awesome APP """


@app.command(name='list')
def get_backups():
    table = Table()

    table.add_column('ID')
    table.add_column('Nombre')
    table.add_column('Base de datos')
    table.add_column('Fecha')

    for _, bdata in BackupsService.list_backups():
        table.add_row(
            f"{bdata.id}",
            f"{bdata.name}.{bdata.ext}",
            bdata.database_name,
            bdata.created_at.strftime('%d/%m/%y %H:%M:%S')
        )

    console = Console()
    console.print(table)


@app.command(name='dump')
@click.option('--name', required=True)
@click.option('--database-type', required=True, type=click.Choice(['postgres', 'mongodb']))
@click.option('--host', required=True)
@click.option('--database', required=True)
@click.option('--username', default=None)
@click.option('--password', default=None)
@click.option('--without-password', default=False)
@click.option('--port', default=None)
@click.option('--compress', default=None, is_flag=True)
@click.option('--format', default=None)
@click.option('--file', required=False, default=None)
def make_backup(name: str, database_type, host, database, username, password, without_password, port, compress, **kwargs):
    name = name.replace('-', '_')

    if without_password is False and password is None:
        password = getpass.getpass('Password: ')

    if without_password is True:
        password = None

    service = BackupsService.build(
        database_type,
        host=host,
        database=database,
        username=username,
        password=password,
        port=port,
    )

    console = Console()
    with console.status("[bold green]Dumping database ..."):
        service.dump(name, compress=compress, **kwargs)
        time.sleep(1)
    console.print("[bold green] Backup created !")


@app.command(name='restore')
@click.option('--database-type', required=True, type=click.Choice(['postgres', 'mongodb']))
@click.option('--host', required=True)
@click.option('--port', default=None)
@click.option('--database', required=True)
@click.option('--username', required=True)
@click.option('--password', default=None)
@click.option('--without-password', default=False)
def restore_backup(database_type, host, port, database, username, password, without_password):
    if without_password is False and password is None:
        password = getpass.getpass('Password: ')

    if without_password is True:
        password = None

    service = BackupsService.build(
        database_type,
        host=host,
        port=port,
        database=database,
        username=username,
        password=password
    )
    console = Console()

    backups = [v for v in BackupsService.list_backups()]
    backups.sort(key=lambda v: v[1].id)
    backups = [
        f"{d.id} - {d.name}.{d.ext} - {d.database_name}" for _, d in backups]

    questions = [
        inquirer.List(
            name='backup',
            message='Select the backup to restore',
            choices=backups
        )
    ]
    answers = inquirer.prompt(questions)

    if not answers:
        raise ValueError('We need than you selected a backup to restore')

    backup_id = int(answers['backup'].split(' - ')[0])
    finded_backup = BackupsService.get_backup(backup_id)

    with console.status("[bold green]Restoring backup ..."):
        service.restore(finded_backup.backup)
        time.sleep(1)
    console.print("[bold green]Backup restored !")


def main():
    build_config()
    app()


if __name__ == '__main__':
    main()



from pathlib import Path
from .base import Backup
import subprocess
import os
from io import BytesIO

DUMP_SQL_FORMAT = 'p'
DUMP_CUSTOM_FORMAT = 'c'
DUMP_DIRECTORY_FORMAT = 'd'
DUMP_TAR_FORMAT = 't'
DUMP_DEFAULT_FORMAT = 'p'


class Postgres(Backup):

    CMDS_TO_CHECK = [
        ('pg_dump', '--version'),
        ('pg_restore', '--version'),
        ('psql', '--version')]

    def dump(self, **kwargs) -> BytesIO:
        env = os.environ.copy()
        if self.password:
            env['PGPASSWORD'] = self.password

        config = {
            "format": kwargs.get('format', DUMP_DEFAULT_FORMAT) or DUMP_DEFAULT_FORMAT,
            "compress": kwargs.get('compress', False),
            "compress_level": kwargs.get('compress_level', 9),
            "file": kwargs.get('file', None)
        }

        config_list = [
            "-F",
            config['format'],
            "-Z",
            str(config['compress_level'] if config['compress'] else 0),
        ]

        if not config['format'] == DUMP_SQL_FORMAT and not config['format'] == DUMP_CUSTOM_FORMAT and not config['format'] == DUMP_TAR_FORMAT:
            config_list.append('-f')
            if not config['file']:
                raise ValueError('The output format need file kwarg')
            config_list.append(str(Path(config['file']).resolve()))

        p = subprocess.Popen([
            'pg_dump',
            '-h',
            self.host,
            '-U',
            self.username,
            '-p',
            f"{self.port}",
            self.database,
            '-w' if self.password else '--no-password',
            *config_list
        ], env=env, stdout=subprocess.PIPE)

        code = p.wait()
        if code > 0:
            raise Exception(p.stderr.read().decode('utf-8'))

        return BytesIO(p.stdout.read())

    def restore(self, file_path: Path) -> BytesIO:
        env = os.environ.copy()
        if self.password:
            env['PGPASSWORD'] = self.password

        args = [
            '-h',
            self.host,
            '-p',
            f"{self.port}",
            '-U',
            self.username,
            '-d',
            self.database,
        ]

        f_arg = [
            '-f',
            f"{file_path.absolute()}",
        ] if str(file_path).endswith('.sql') else [
            f"{file_path.absolute()}",
        ]

        p = subprocess.Popen([
            'psql' if str(file_path).endswith('.sql') else 'pg_restore',
            *args,
            *f_arg
        ], env=env, stdout=subprocess.PIPE, stderr=subprocess.PIPE)

        code = p.wait()
        if code > 0:
            print(p.stderr)
            raise Exception(p.stderr.read().decode('utf-8'))



from datetime import datetime
from dataclasses import dataclass, field
from typing import NamedTuple
from pathlib import Path


@dataclass
class BackupData:
    id: int
    name: str
    ext: str
    backup: Path
    database_name: str
    created_at: datetime = field(default_factory=lambda: datetime.now())


class BackupFile(NamedTuple):
    id: int
    name: str
    ext: str
    file: Path



from pydbbackups.cli.utils import get_backup_class, save_backup_file, get_backups_data, get_backups_files, DTEncoder, mongo_ext_formatter, postgres_ext_formatter
from pydbbackups.cli.config import BACKUPS_DATA_DIR
from pydbbackups.cli.models import BackupData, BackupFile

from pydbbackups import Backup
from datetime import datetime
from dataclasses import asdict
import json
from pathlib import Path


def backup_ext_formatter(db_type: str, name: str, **kwargs) -> str:
    databases = {
        "postgres": postgres_ext_formatter,
        "mongodb": mongo_ext_formatter
    }

    formatter = databases.get(db_type, None)

    if not formatter:
        return name

    return formatter(name, **kwargs)


class BackupsService:
    backup_cls: Backup
    db_type: str

    def __init__(self, backup_cls: Backup = None) -> None:
        self.backup_cls = backup_cls

    @staticmethod
    def build(db_type: str, **kwargs):
        backup_cls = get_backup_class(db_type)(**kwargs)
        cls = BackupsService(backup_cls)
        cls.db_type = db_type

        return cls

    @staticmethod
    def list_backups():
        data = get_backups_data()
        backups = get_backups_files()

        for bfile in backups:
            # Get the first object
            bdata = next((obj for obj in data if obj.id == bfile.id), None)
            if not bdata:
                continue

            yield bfile, bdata

    @staticmethod
    def get_backup(backup_id: int):
        data = get_backups_data()
        for v in data:
            if v.id == backup_id:
                return v

    def dump(self, name: str, **kwargs):
        output = self.backup_cls.dump(**kwargs)
        content = output.read()

        data = get_backups_data()
        created_id = len(data)
        created_at = datetime.now()

        name = backup_ext_formatter(
            self.db_type, name, **kwargs)

        if len(name.split('.')) > 1:
            [name, ext] = name.split('.')

        if not output or len(content) == 0:
            print('Warning: This backup will not be saved')
            return
        else:
            meta = save_backup_file(BackupFile(
                id=created_id, name=name, ext=ext, file=Path()), content)
            print(meta.file, 'Created !')

        data.append(BackupData(**{
            "id": len(data),
            "name": meta.name,
            "ext": meta.ext,
            "database_name": self.db_type,
            "backup": meta.file,
            "created_at": created_at
        }))

        serialized_data = [asdict(v) for v in data]
        BACKUPS_DATA_DIR.write_text(json.dumps(serialized_data, cls=DTEncoder))

    def restore(self, backup: Path):
        self.backup_cls.restore(backup)
